#
# Copyright (c) 2011 Yubico AB
# See the file COPYING for licence statement.
#
"""
Get random data from TRNG on YubiHSM and insert it into host
entropy pool. Probably only works on Linux since the ioctl()
request value RNDADDENTROPY seems Linux specific.
"""

import os
import sys
import fcntl
import struct
import argparse
import pyhsm

default_device = "/dev/ttyACM0"
default_iterations = 100
default_entropy_ratio = 2	# number of bits of entropy per byte of random data

RNDADDENTROPY = 1074287107	# from /usr/include/linux/random.h

def parse_args():
    """
    Parse the command line arguments
    """
    parser = argparse.ArgumentParser(description = "Add random data from YubiHSM to Linux entropy",
                                     add_help = True,
                                     formatter_class = argparse.ArgumentDefaultsHelpFormatter,
                                     )
    parser.add_argument('-D', '--device',
                        dest='device',
                        default=default_device,
                        required=False,
                        help='YubiHSM device',
                        )
    parser.add_argument('-v', '--verbose',
                        dest='verbose',
                        action='store_true', default=False,
                        help='Enable verbose operation'
                        )
    parser.add_argument('--debug',
                        dest='debug',
                        action='store_true', default=False,
                        help='Enable debug operation'
                        )
    parser.add_argument('-r', '--ratio',
                        dest='ratio',
                        type=int, default=default_entropy_ratio,
                        help='Bits per byte of data read to use as entropy',
                        )
    parser.add_argument('-c', '--count',
                        dest='iterations',
                        type=int, default=default_iterations,
                        help='Number of iterations to run',
                        )

    args = parser.parse_args()

    return args


def get_entropy(hsm, iterations, entropy_ratio):
    """
    Read entropy from YubiHSM and feed it to Linux as entropy using ioctl() syscall.
    """
    fd = os.open("/dev/random", os.O_WRONLY)
    # struct rand_pool_info {
    #    int     entropy_count;
    #    int     buf_size;
    #    __u32   buf[0];
    # };
    fmt = 'ii%is' % (pyhsm.defines.YSM_MAX_PKT_SIZE - 1)
    for _ in xrange(iterations):
        rnd = hsm.random(pyhsm.defines.YSM_MAX_PKT_SIZE - 1)
        this = struct.pack(fmt, entropy_ratio * len(rnd), len(rnd), rnd)
        fcntl.ioctl(fd, RNDADDENTROPY, this)
    os.close(fd)

def main():
    """
    What will be executed when running as a stand alone program.
    """
    args = parse_args()

    try:
        s = pyhsm.base.YHSM(device=args.device, debug=args.debug)
        get_entropy(s, args.iterations, args.ratio)
        return 0
    except pyhsm.exception.YHSM_Error as e:
        sys.stderr.write("ERROR: %s" % (e.reason))
        return 1



#
# Copyright (C) 2012-2013 Yubico AB. All rights reserved.
#
"""
This is a tool to decrypt AEADs generated using a YubiHSM, provided that
you know the key_handle used as well as the AES key used.

This can be used together with yhsm-generate-keys to generate a number
of AEADs, and then decrypt them to program YubiKeys accordingly.
"""

import os
import re
import sys
import fcntl
import argparse
import traceback

import pyhsm

args = None
yknum = 0

def parse_args():
    """
    Parse the command line arguments
    """
    parser = argparse.ArgumentParser(description = 'Decrypt AEADs',
                                     add_help = True,
                                     formatter_class = argparse.ArgumentDefaultsHelpFormatter,
                                     )
    parser.add_argument('-v', '--verbose',
                        dest='verbose',
                        action='store_true', default=False,
                        help='Enable verbose operation',
                        )
    parser.add_argument('--debug',
                        dest='debug',
                        action='store_true', default=False,
                        help='Enable debug operation',
                        )
    parser.add_argument('--format',
                        dest='format',
                        default='raw',
                        help='Select output format (aead, raw or yubikey-csv)',
                        )
    parser.add_argument('--output-dir',
                        dest='output_dir',
                        help='Output dir basename (for --format aead)',
                        metavar='DIR',
                        )
    parser.add_argument('--print-filename',
                        dest='print_filename',
                        action='store_true', default=False,
                        help='Prefix each row with the AEAD filename',
                        )
    parser.add_argument('--key-handle',
                        dest='key_handle',
                        help='Key handle used when generating the AEADs.',
                        metavar='HANDLE',
                        )
    parser.add_argument('--key-handle-out',
                        dest='key_handle_out',
                        help='Key handle used when generating *new* AEADs (with --format aead).',
                        metavar='HANDLE',
                        )
    parser.add_argument('--aes-key',
                        dest='aes_key',
                        required=True,
                        help='AES key used when generating the AEADs.',
                        metavar='HEXSTR',
                        )
    parser.add_argument('--aes-key-out',
                        dest='aes_key_out',
                        required=False,
                        help='AES key used when generating *new* AEADs (with --format aead).',
                        metavar='HEXSTR',
                        )
    parser.add_argument('--start-public-id',
                        dest='start_id',
                        required=False, default=None,
                        help='The first public id to decrypt',
                        metavar='INT-OR-MODHEX',
                        )
    parser.add_argument('--stop-public-id',
                        dest='stop_id',
                        required=False, default=None,
                        help='The last public id to decrypt',
                        metavar='INT-OR-MODHEX',
                        )
    parser.add_argument('--fail-fast',
                        dest='fail_fast',
                        action='store_true', default=False,
                        help='Terminate on the first AEAD failure, rather than keep going.',
                        )
    parser.add_argument('paths',
                        nargs='+',
                        help='Files and/or directories to process.',
                        metavar='FILE-OR-DIR'
                        )
    args = parser.parse_args()
    # argument fixups
    args.format = args.format.lower()
    args.aes_key = args.aes_key.decode('hex')
    if args.key_handle:
        args.key_handle = pyhsm.util.key_handle_to_int(args.key_handle)
    if args.start_id is not None:
        try:
            n = int(args.start_id)
        except ValueError:
            hexstr = pyhsm.yubikey.modhex_decode(args.start_id)
            n = int(hexstr, 16)
        args.start_id = n
    if args.stop_id is not None:
        try:
            n = int(args.stop_id)
        except ValueError:
            hexstr = pyhsm.yubikey.modhex_decode(args.stop_id)
            n = int(hexstr, 16)
        args.stop_id = n
    # some checks
    if args.format == 'aead':
        if not args.output_dir:
            sys.stderr.write("error: --output-dir is required when using --format aead.\n")
            return False
        if not os.path.isdir(args.output_dir):
            sys.stderr.write("error: Output directory '%s' not found\n" % (args.output_dir))
            return False
        if not args.aes_key_out:
            sys.stderr.write("error: --aes-key-out is required when using --format aead.\n")
            return False
        if not args.key_handle_out:
            sys.stderr.write("error: --key-handle-out is required when using --format aead.\n")
            return False
        # argument fixups
        args.aes_key_out = args.aes_key_out.decode('hex')
        args.key_handle_out_orig = args.key_handle_out # save to use in AEAD output paths
        args.key_handle_out = pyhsm.util.key_handle_to_int(args.key_handle_out)
    return args

class MyState():
    """
    Class to keep track of failed files.
    """
    def __init__(self, args):
        self.args = args
        self.failed_files = []
        self.file_count = 0

    def log_failed(self, fn):
        self.failed_files.append(fn)
        self.file_count += 1

    def log_success(self, fn):
        self.file_count += 1

    def should_quit(self):
        return (self.failed_files and self.args.fail_fast)

def process_file(path, fn, args, state):
    """
    The main function for reading a file and decrypting it.
    """
    full_fn = os.path.join(path, fn)

    if not re.match("^[cbdefghijklnrtuv]+$", fn):
        if args.debug:
            sys.stderr.write("warning: Ignoring non-modhex file '%s'\n" % (full_fn))
        return True

    if (args.start_id is not None) or (args.stop_id is not None):
        this = int(pyhsm.yubikey.modhex_decode(fn), 16)
        if (args.start_id is not None) and this < args.start_id:
            if args.debug:
                sys.stderr.write("warning: Skipping public id %s (%i) < %i\n" % (fn, this, args.start_id))
            return True
        if (args.stop_id is not None) and this > args.stop_id:
            if args.debug:
                sys.stderr.write("warning: Skipping public id %s (%i) > %i\n" % (fn, this, args.stop_id))
            return True

    if args.debug:
        sys.stderr.write("Loading AEAD : %s\n" % full_fn)

    aead = pyhsm.aead_cmd.YHSM_GeneratedAEAD(None, None, '')
    aead.load(full_fn)

    if not aead.nonce:
        # AEAD file version 0, need to fill in nonce etc.
        if args.key_handle is None:
            sys.stderr.write("error: AEAD in file %s does not include key_handle, and none provided.\n" % (full_fn))
            state.log_failed(full_fn)
            return False
        aead.key_handle = args.key_handle
        aead.nonce = pyhsm.yubikey.modhex_decode(fn).decode('hex')

    if args.debug:
        sys.stderr.write("%s\n" % aead)
        sys.stderr.write("AEAD len %i : %s\n" % (len(aead.data), aead.data.encode('hex')))
    pt = pyhsm.soft_hsm.aesCCM(args.aes_key, aead.key_handle, aead.nonce, aead.data, decrypt = True)

    if args.print_filename:
        print("%s " % (full_fn)),

    if args.format == 'raw':
        print(pt.encode('hex'))
    elif args.format == 'aead':
        # encrypt secrets with new key
        ct = pyhsm.soft_hsm.aesCCM(args.aes_key_out, args.key_handle_out, aead.nonce, pt, decrypt = False)
        aead_out = pyhsm.aead_cmd.YHSM_GeneratedAEAD(aead.nonce, args.key_handle_out, ct)
        filename = aead_filename(args.output_dir, args.key_handle_out_orig, fn)
        aead_out.save(filename)
        if args.print_filename:
            print ""
    elif args.format == 'yubikey-csv':
        key = pt[:pyhsm.defines.KEY_SIZE]
        uid = pt[pyhsm.defines.KEY_SIZE:]
        access_code = '00' * 6
        timestamp = ''
        global yknum
        yknum += 1
        print("%i,%s,%s,%s,%s,%s,,,,," % (yknum,
                                          fn,
                                          uid.encode('hex'),
                                          key.encode('hex'),
                                          access_code,
                                          timestamp,
                                          ))

    state.log_success(full_fn)

    return True

def aead_filename(aead_dir, key_handle, public_id):
    """
    Return the filename of the AEAD for this public_id,
    and create any missing directorys.
    """
    parts = [aead_dir, key_handle] + pyhsm.util.group(public_id, 2)
    path = os.path.join(*parts)

    if not os.path.isdir(path):
        os.makedirs(path)

    return os.path.join(path, public_id)

def safe_process_files(path, files, args, state):
    """
    Process a number of files in a directory. Catches any exception from the
    processing and checks if we should fail directly or keep going.
    """
    for fn in files:
        full_fn = os.path.join(path, fn)
        try:
            if not process_file(path, fn, args, state):
                return False
        except Exception, e:
            sys.stderr.write("error: %s\n%s\n" % (os.path.join(path, fn), traceback.format_exc()))
            state.log_failed(full_fn)
        if state.should_quit():
            return False
    return True

def walk_dir(path, args, state):
    """
    Check all files in `path' to see if there is any requests that
    we should send out on the bus.
    """
    if args.debug:
        sys.stderr.write("Walking %s\n" % path)

    for root, _dirs, files in os.walk(path):
        if not safe_process_files(root, files, args, state):
            return False
        if state.should_quit():
            return False
    return True

def main():
    """ Main function when running as a program. """
    global args
    args = parse_args()

    if not args:
        return 1

    state = MyState(args)

    for path in args.paths:
        if os.path.isdir(path):
            walk_dir(path, args, state)
        else:
            safe_process_files(os.path.dirname(path), [os.path.basename(path)], args, state)
        if state.should_quit():
            break
    if state.failed_files:
        sys.stderr.write("error: %i/%i AEADs failed\n" % (len(state.failed_files), state.file_count))
        return 1
    if args.debug:
        sys.stderr.write("Successfully processed %i AEADs\n" % (state.file_count))



#
# Copyright (c) 2011-2014 Yubico AB
# See the file COPYING for licence statement.
#
"""
Small network server decrypting YubiKey OTPs using an attached YubiHSM.

To support unlimited numbers of YubiKeys, the YubiKey AES keys are
stored in AEAD's (Authenticated Encryption with Associated Data) on
the host computer.

When an OTP is received, we find the right AEAD for this key (based on
the public ID of the YubiKey), and then send the AEAD together with the
OTP to the YubiHSM. The YubiHSM is able to decrypt the AEAD (if it has
the appropriate key handle configured), and then able to decrypt the
YubiKey OTP using the AES key stored in the AEAD.

The information the YubiKey encrypted using it's AES key is then
returned in clear text from the YubiHSM. This includes the counter
information and also (relative) timestamps.

It is not the job of the KSM (or YubiHSM) to ensure that the OTP has
not been seen before - that is done by the validation server (using
the database) :

     O            +----------+
    /|\           |Validation|     +-----+   +---------+
     |  -- OTP--> |  server  | --> | KSM +---| YubiHSM |
    / \           +----------+     +-----+   +---------+
                        |
    user             +--+--+
                     | DB  |
                     +-----+
"""

import os
import sys
import BaseHTTPServer
import socket
import argparse
import syslog
import re
import pyhsm
import pyhsm.yubikey
import serial
import daemon
import sqlalchemy
from functools import partial
from pyhsm.soft_hsm import SoftYHSM

default_device = "/dev/ttyACM0"
default_dir = "/var/cache/yubikey-ksm/aeads"
default_serve_url = "/wsapi/decrypt?otp="
default_listen_addr = "127.0.0.1"
default_port = 8002
default_reqtimeout = 5
default_pid_file = None
default_db_url = None

valid_input_from_key = re.compile('^[cbdefghijklnrtuv]{32,48}$')


stats = { 'ok': 0,
    'invalid': 0,
    'no_aead': 0,
    'err': 0 }

context = daemon.DaemonContext()


class YHSM_KSMRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    """
    Handle a HTTP request.

    Try to be careful and validate the input, and then look for an AEAD file matching the
    public id of the OTP. If an AEAD for one of our key handles is found, we ask the YubiHSM
    to decrypt the OTP using the AEAD and return the result (counter and timestamp information).
    """

    def __init__(self, hsm, aead_backend, args, *other_args, **kwargs):
        self.hsm = hsm
        self.verbose = args.debug or args.verbose

        self.serve_url = args.serve_url
        self.stats_url = args.stats_url
        self.key_handles = args.key_handles
        self.timeout = args.reqtimeout
        self.aead_backend = aead_backend
        BaseHTTPServer.BaseHTTPRequestHandler.__init__(self, *other_args, **kwargs)

    def do_GET(self):
        """ Handle a HTTP GET request. """
        # Example session:
        # in  : GET /wsapi/decrypt?otp=ftftftccccdvvbfcfduvvcubikngtchlubtutucrld HTTP/1.0
        # out : OK counter=0004 low=f585 high=3e use=03
        if self.path.startswith(self.serve_url):
            from_key = self.path[len(self.serve_url):]

            val_res = self.decrypt_yubikey_otp(from_key)

            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            self.wfile.write(val_res)
            self.wfile.write("\n")
        elif self.stats_url and self.path == self.stats_url:
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            for key in stats:
                self.wfile.write("%s %d\n" % (key, stats[key]))
        else:
            self.log_error("Bad URL '%s' - I'm serving '%s' (responding 403)" % (self.path, self.serve_url))
            self.send_response(403, 'Forbidden')
            self.end_headers()

    def decrypt_yubikey_otp(self, from_key):
        """
        Try to decrypt a YubiKey OTP.

        Returns a string starting with either 'OK' or 'ERR' :

        'OK counter=ab12 low=dd34 high=2a use=0a'

        'ERR Unknown public_id'

        on YubiHSM errors (or bad OTP), only 'ERR' is returned.
        """
        if not re.match(valid_input_from_key, from_key):
            self.log_error("IN: %s, Invalid OTP" % (from_key))
            if self.stats_url:
                stats['invalid'] += 1
            return "ERR Invalid OTP"

        public_id, _otp = pyhsm.yubikey.split_id_otp(from_key)

        try:
            aead = self.aead_backend.load_aead(public_id)
        except Exception as e:
            self.log_error(str(e))
            if self.stats_url:
                stats['no_aead'] += 1
            return "ERR Unknown public_id"

        try:
            res = pyhsm.yubikey.validate_yubikey_with_aead(
                self.hsm, from_key, aead, aead.key_handle)
            # XXX double-check public_id in res, in case BaseHTTPServer suddenly becomes multi-threaded
            # XXX fix use vs session counter confusion
            val_res = "OK counter=%04x low=%04x high=%02x use=%02x" % \
                (res.use_ctr, res.ts_low, res.ts_high, res.session_ctr)
            if self.stats_url:
                stats['ok'] += 1
        except pyhsm.exception.YHSM_Error as e:
            self.log_error ("IN: %s, Validate FAILED: %s" % (from_key, str(e)))
            val_res = "ERR"
            if self.stats_url:
                stats['err'] += 1

        self.log_message("SUCCESS OTP %s PT hsm %s", from_key, val_res)
        return val_res

    def log_error(self, fmt, *fmt_args):
        """ Log to syslog. """
        msg = self.my_address_string() + " - - " + fmt % fmt_args
        my_log_message(self.verbose, syslog.LOG_ERR, msg)

    def log_message(self, fmt, *fmt_args):
        """ Log to syslog. """
        msg = self.my_address_string() + " - - " + fmt % fmt_args
        my_log_message(self.verbose, syslog.LOG_INFO, msg)

    def my_address_string(self):
        """ For logging client host without resolving. """
        return getattr(self, 'client_address', ('', None))[0]


class FSBackend(object):

    def __init__(self, aead_dir, key_handles):
        self.aead_dir = aead_dir
        self.key_handles = key_handles
        if not os.path.isdir(aead_dir):
            raise ValueError("AEAD directory '%s' does not exist." % aead_dir)

    def load_aead(self, public_id):
        fn_list = []
        for kh, kh_int in self.key_handles:
            aead = pyhsm.aead_cmd.YHSM_GeneratedAEAD(None, kh_int, '')
            filename = aead_filename(self.aead_dir, kh, public_id)
            fn_list.append(filename)
            try:
                aead.load(filename)
                if not aead.nonce:
                    aead.nonce = pyhsm.yubikey.modhex_decode(public_id).decode('hex')
                return aead
            except IOError:
                continue
        raise Exception("Attempted to load AEAD from : %s" % (fn_list))


class SQLBackend(object):
    def __init__(self, db_url):
        self.engine = sqlalchemy.create_engine(db_url)
        metadata = sqlalchemy.MetaData()
        self.aead_table = sqlalchemy.Table('aead_table', metadata, autoload=True, autoload_with=self.engine)

    def load_aead(self, public_id):
        """ Loads AEAD from the specified database. """
        connection = self.engine.connect()
        trans = connection.begin()

        try:
            s = sqlalchemy.select([self.aead_table]).where(self.aead_table.c.public_id == public_id)
            result = connection.execute(s)

            for row in result:
                kh_int = row['keyhandle']
                aead = pyhsm.aead_cmd.YHSM_GeneratedAEAD(None, kh_int, '')
                aead.data = row['aead']
                aead.nonce = row['nonce']
            return aead
        except Exception:
            trans.rollback()
            raise Exception("No AEAD in DB for public_id %s (%s)" % (public_id, str(e)))
        finally:
            connection.close()


class YHSM_KSMServer(BaseHTTPServer.HTTPServer):
    """
    Wrapper class to properly initialize address_family for IPv6 addresses.
    """
    def __init__(self, server_address, req_handler):
        if ":" in server_address[0]:
            self.address_family = socket.AF_INET6
        BaseHTTPServer.HTTPServer.__init__(self, server_address, req_handler)


def aead_filename(aead_dir, key_handle, public_id):
    """
    Return the filename of the AEAD for this public_id.
    """
    parts = [aead_dir, key_handle] + pyhsm.util.group(public_id, 2) + [public_id]
    return os.path.join(*parts)


def parse_args():
    """
    Parse the command line arguments
    """
    parser = argparse.ArgumentParser(description="Decrypt YubiKey OTPs using YubiHSM",
                                     add_help=True,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter
                                     )
    parser.add_argument('-D', '--device',
                        dest='device',
                        default=default_device,
                        required=False,
                        help='YubiHSM device'
                        )
    parser.add_argument('-B', '--aead-dir',
                        dest='aead_dir',
                        default=default_dir,
                        required=False,
                        help='AEAD directory - base directory of your AEADs',
                        metavar='DIR',
                        )
    parser.add_argument('-U', '--serve-url',
                        dest='serve_url',
                        default=default_serve_url,
                        required=False,
                        help='Base URL for decrypt web service',
                        metavar='URL',
                        )
    parser.add_argument('-S', '--stats-url',
                        dest='stats_url',
                        required=False,
                        help='URL where statistics can be retrieved',
                        metavar='URL',
                        )
    parser.add_argument('-v', '--verbose',
                        dest='verbose',
                        action='store_true', default=False,
                        help='Enable verbose operation'
                        )
    parser.add_argument('-d', '--daemon',
                        dest='daemon',
                        action='store_true', default=False,
                        help='Run as daemon'
                        )
    parser.add_argument('--debug',
                        dest='debug',
                        action='store_true', default=False,
                        help='Enable debug operation'
                        )
    parser.add_argument('--port',
                        dest='listen_port',
                        type=int, default=default_port,
                        required=False,
                        help='Port to listen on',
                        metavar='PORT',
                        )
    parser.add_argument('--addr',
                        dest='listen_addr',
                        default=default_listen_addr,
                        required=False,
                        help='Address to bind to',
                        metavar='ADDR',
                        )
    parser.add_argument('--reqtimeout',
                        dest='reqtimeout',
                        type=int, default=default_reqtimeout,
                        required=False,
                        help='Request timeout in seconds',
                        metavar='SECONDS',
                        )
    parser.add_argument('--key-handle', '--key-handles',
                        dest='key_handles',
                        nargs='+',
                        required=True,
                        help='Key handle(s) to use to decrypt AEADs on the YHSM.',
                        metavar='HANDLE',
                        )
    parser.add_argument('--pid-file',
                        dest='pid_file',
                        default=default_pid_file,
                        required=False,
                        help='PID file',
                        metavar='FILENAME',
                        )
    parser.add_argument('--db-url',
                        dest='db_url',
                        default=default_db_url,
                        required=False,
                        help='The database url to read the AEADs from, you can '
                        'also use env:YOURENVVAR to instead read the URL from '
                        'the YOURENVVAR environment variable.',
                        metavar='DBURL',
                        )

    return parser.parse_args()


def args_fixup(args):
    """
    Additional checks/cleanups of parsed arguments.
    """
    # cache key_handle_to_int of all key handles, turning args.key_handles into
    # a list of tuples with both original value and integer
    res = []
    for kh in args.key_handles:
        kh_int = pyhsm.util.key_handle_to_int(kh)
        res.append((kh, kh_int,))
    args.key_handles = res

    # Check if the DB url should be read from an environment variable
    if args.db_url and args.db_url.startswith('env:'):
        env_var = args.db_url[4:]
        if env_var in os.environ:
            args.db_url = os.environ[env_var]


def write_pid_file(fn):
    """ Create a file with our PID. """
    if not fn:
        return None
    if fn == '' or fn == "''":
        # work around argument passings in init-scripts
        return None
    f = open(fn, "w")
    f.write("%s\n" % (os.getpid()))
    f.close()


def run(hsm, aead_backend, args):
    """
    Start a BaseHTTPServer.HTTPServer and serve requests forever.
    """

    write_pid_file(args.pid_file)

    server_address = (args.listen_addr, args.listen_port)
    httpd = YHSM_KSMServer(server_address,
                           partial(YHSM_KSMRequestHandler, hsm, aead_backend, args))
    my_log_message(args.debug or args.verbose, syslog.LOG_INFO,
                   "Serving requests to 'http://%s:%s%s' with key handle(s) %s (YubiHSM: '%s', AEADs in '%s', DB in '%s')"
                   % (args.listen_addr, args.listen_port, args.serve_url, args.key_handles, args.device, args.aead_dir, args.db_url))
    httpd.serve_forever()


def my_log_message(verbose, prio, msg):
    """
    Log to syslog, and possibly also to stderr.
    """
    syslog.syslog(prio, msg)
    if verbose or prio == syslog.LOG_ERR:
        sys.stderr.write("%s\n" % (msg))


def main():
    """
    Main program.
    """
    my_name = os.path.basename(sys.argv[0])
    if not my_name:
        my_name = "yhsm-yubikey-ksm"
    syslog.openlog(my_name, syslog.LOG_PID, syslog.LOG_LOCAL0)

    args = parse_args()
    args_fixup(args)

    aead_backend = None
    if args.db_url:
        # Using an SQL database for AEADs
        try:
            aead_backend = SQLBackend(args.db_url)
        except Exception as e:
            my_log_message(args.debug or args.verbose, syslog.LOG_ERR,
                           'Could not connect to database "%s" : %s' % (args.db_url, e))
            return 1
    else:
        # Using the filesystem for AEADs
        try:
            aead_backend = FSBackend(args.aead_dir, args.key_handles)
        except Exception as e:
            my_log_message(args.debug or args.verbose, syslog.LOG_ERR,
                           'Could not create AEAD FSBackend: %s' % e)
            return 1

    if args.device == '-':
        # Using a soft-HSM with keys from stdin
        try:
            hsm = SoftYHSM.from_json(sys.stdin.read(), debug=args.debug)
        except ValueError as e:
            my_log_message(args.debug or args.verbose, syslog.LOG_ERR,
                           'Failed opening soft YHSM from stdin : %s' % (e))
            return 1
    elif os.path.isfile(args.device):
        # Using a soft-HSM from file
        try:
            hsm = SoftYHSM.from_file(args.device, debug=args.debug)
        except ValueError as e:
            my_log_message(args.debug or args.verbose, syslog.LOG_ERR,
                           'Failed opening soft YHSM "%s" : %s' % (args.device, e))
            return 1
    else:
        # Using a real HSM
        try:
            hsm = pyhsm.YHSM(device=args.device, debug=args.debug)
            context.files_preserve = [hsm.get_raw_device()]
        except serial.SerialException as e:
            my_log_message(args.debug or args.verbose, syslog.LOG_ERR,
                           'Failed opening YubiHSM device "%s" : %s' % (args.device, e))
            return 1

    if args.daemon:
        with context:
            run(hsm, aead_backend, args)
    else:
        try:
            run(hsm, aead_backend, args)
        except KeyboardInterrupt:
            print ""
            print "Shutting down"
            print ""


if __name__ == '__main__':
    sys.exit(main())



#
# Copyright (c) 2011 Yubico AB
# See the file COPYING for licence statement.
#
"""
Utility to unlock the key store of a YubiHSM,
using the 'HSM password'/'master key'.
"""

import sys
import pyhsm
import argparse
import getpass

default_device = "/dev/ttyACM0"

def parse_args():
    """
    Parse the command line arguments
    """
    parser = argparse.ArgumentParser(description = "Unlock key store of YubiHSM",
                                     add_help = True,
                                     formatter_class = argparse.ArgumentDefaultsHelpFormatter,
                                     )
    parser.add_argument('-D', '--device',
                        dest='device',
                        default=default_device,
                        required=False,
                        help='YubiHSM device',
                        )
    parser.add_argument('-v', '--verbose',
                        dest='verbose',
                        action='store_true', default=False,
                        help='Enable verbose operation'
                        )
    parser.add_argument('--debug',
                        dest='debug',
                        action='store_true', default=False,
                        help='Enable debug operation'
                        )
    parser.add_argument('--no-otp',
                        dest='no_otp',
                        action='store_true', default=False,
                        help='Don\'t ask for OTP, even if YubiHSM supports it',
                        )
    parser.add_argument('--stdin',
                        dest='stdin',
                        action='store_true', default=False,
                        help='Read data from stdin instead of prompting',
                        )

    args = parser.parse_args()

    return args

def get_password(hsm, args):
    """ Get password of correct length for this YubiHSM version. """
    expected_len = 32
    name = 'HSM password'
    if hsm.version.have_key_store_decrypt():
        expected_len = 64
        name = 'master key'

    if args.stdin:
        password = sys.stdin.readline()
        while password and password[-1] == '\n':
            password = password[:-1]
    else:
        if args.debug:
            password = raw_input('Enter %s (press enter to skip) (will be echoed) : ' % (name))
        else:
            password = getpass.getpass('Enter %s (press enter to skip) : ' % (name))

    if len(password) <= expected_len:
        password = password.decode('hex')
        if not password:
            return None
        return password
    else:
        sys.stderr.write("ERROR: Invalid HSM password (expected max %i chars, got %i)\n" % \
                              (expected_len, len(password)))
        return 1

def get_otp(hsm, args):
    """ Get OTP from YubiKey. """
    if args.no_otp:
        return None
    if hsm.version.have_unlock():
        if args.stdin:
            otp = sys.stdin.readline()
            while otp and otp[-1] == '\n':
                otp = otp[:-1]
        else:
            otp = raw_input('Enter admin YubiKey OTP (press enter to skip) : ')
        if len(otp) == 44:
            # YubiHSM admin OTP's always have a public_id length of 6 bytes
            return otp
        if otp:
            sys.stderr.write("ERROR: Invalid YubiKey OTP\n")
    return None

def main():
    """
    What will be executed when running as a stand alone program.
    """
    args = parse_args()

    try:
        hsm = pyhsm.base.YHSM(device=args.device, debug=args.debug)

        if args.debug or args.verbose:
            print "Device  : %s" % (args.device)
            print "Version : %s" % (hsm.info())
            print ""

        password = get_password(hsm, args)
        otp = get_otp(hsm, args)
        if not password and not otp:
            print "\nAborted\n"
            return 1
        else:
            if args.debug or args.verbose:
                print ""
            if hsm.unlock(password = password, otp = otp):
                if args.debug or args.verbose:
                    print "OK\n"
    except pyhsm.exception.YHSM_Error, e:
        sys.stderr.write("ERROR: %s\n" % (e.reason))
        return 1

    return 0



#
# Copyright (c) 2011, 2012 Yubico AB
# See the file COPYING for licence statement.
#
"""
 Credential validation server utilizing YubiHSM.

 Modes of operation :

  OTP - YubiKey validation using internal DB in YubiHSM.
        The YubiHSM can take care of complete Yubico OTP
        validation - including storing seen counter values
        in an internal database.

        There is an --otp mode available that tries to be
        compatible with YK-VAL, and there is a --short-otp
        mode that gives responses looking like YK-KSM.

  HOTP - OATH-HOTP validation using secrets stored on host
         computer (in secure AEADs only decryptable inside
         YubiHSM). The HMAC-SHA1 of the OATH counter value
         is done inside the YubiHSM, so the OATH Key is
         never exposed outside the YubiHSM.

  TOTP - OATH-TOTP validation using secrets stored on host
         computer (in secure AEADs only decryptable inside
         YubiHSM). The HMAC-SHA1 of the OATH timecounter value
         is compared, so the OATH Key is never exposed
         outside the YubiHSM.

  PWHASH - Uses AEAD plaintext compare in the YubiHSM to see
           if a supplied password hash matches the password
           hash used in an earlier 'set' operation. These
           AEADs can be generated using
           `yhsm-password-auth.py --set ...'.

 All these modes must be explicitly enabled on the command
 line to be allowed (--otp, --hotp, --totp and --pwhash).

 Examples using OATH-HOTP :

   > GET /yhsm/validate?hotp=ubftcdcdckcf359152 HTTP/1.1
   ...
   < HTTP/1.0 200 OK
   < OK counter=0003

   same again (replay), differently formatted :

   > GET /yhsm/validate?uid=ubftcdcdckcf&hotp=359152 HTTP/1.1
   ...
   < HTTP/1.0 200 OK
   < ERR Could not validate OATH-HOTP OTP

 Examples using OATH-TOTP :

   > GET /yhsm/validate?totp=ubftcdcdckcf216781 HTTP/1.1
   ...
   < HTTP/1.0 200 OK
   < OK timecounter=2ed5376

   same again (but outside of time tolerance), differently formatted :

   > GET /yhsm/validate?uid=ubftcdcdckcf&totp=359152 HTTP/1.1
   ...
   < HTTP/1.0 200 OK
   < ERR Could not validate OATH-TOTP OTP


 Example PWHASH (AEAD and NONCE as returned by
                 `yhsm-password-auth.py --set ...') :

 > GET /yhsm/validate?pwhash=pbkdf2-of-password-here&aead=2b70...2257&nonce=010203040506&kh=8192 HTTP/1.1
 ...
 < HTTP/1.0 200 OK
 < OK pwhash validated

"""

import re
import os
import sys
import time
import hmac
import syslog
import serial
import socket
import base64
import hashlib
import sqlite3
import argparse
import urlparse
import BaseHTTPServer
import pyhsm
import pyhsm.yubikey
import pyhsm.oath_hotp
import pyhsm.oath_totp

default_device = "/dev/ttyACM0"
default_serve_url = "/yhsm/validate?"
default_db_file = "/var/yubico/yhsm-validation-server.db"
default_clients_file = "/var/yubico/yhsm-validation-server_client-id.conf"
default_hotp_window = 5
default_totp_interval = 30
default_totp_tolerance = 1
default_pid_file = None

ykotp_valid_input = re.compile('^[cbdefghijklnrtuv]{32,48}$')
hotp_valid_input = re.compile('^[cbdefghijklnrtuv0-9]{6,20}$')
totp_valid_input = re.compile('^[cbdefghijklnrtuv0-9]{6,20}$')

hsm = None
args = None
saved_key_handle = None

client_ids = {}

class YHSM_VALRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    """
    Handle HTTP GET requests according to configuration in global variable `args'.
    """

    def do_GET(self):
        """
        Process validation GET requests.

        All modes of validation (OTP, OATH and PWHASH) must be explicitly
        enabled in `args' to be allowed.
        """
        if self.path.startswith(args.serve_url):
            res = None
            log_res = None
            mode = None
            params = urlparse.parse_qs(self.path[len(args.serve_url):])
            if "otp" in params:
                if args.mode_short_otp:
                    # YubiKey internal db OTP in KSM mode
                    mode = 'YubiKey OTP (short)'
                    res = validate_yubikey_otp_short(self, params)
                elif args.mode_otp:
                    # YubiKey internal db OTP validation 2.0
                    mode = 'YubiKey OTP'
                    res = validate_yubikey_otp(self, params)
                    #status = [x for x in res.split('\n') if x.startswith("status=")]
                    #if len(status) == 1:
                    #    res = status[0][7:]
                    log_res = '&'.join(res.split('\n'))
                else:
                    res = "ERR 'otp/otp2' disabled"
            elif "hotp" in params:
                if args.mode_hotp:
                    mode = 'OATH-HOTP'
                    res = validate_oath_hotp(self, params)
                else:
                    res = "ERR 'hotp' disabled"
            elif "totp" in params:
                if args.mode_totp:
                    mode = 'OATH-TOTP'
                    res = validate_oath_totp(self, params)
                else:
                    res = "ERR 'totp' disabled"
            elif "pwhash" in params:
                if args.mode_pwhash:
                    mode = 'Password hash'
                    res = validate_pwhash(self, params)
                else:
                    res = "ERR 'pwhash' disabled"

            if not log_res:
                log_res = res

            self.log_message("%s validation result: %s -> %s", mode, self.path, log_res)

            if res != None:
                self.send_response(200)
                self.send_header('Content-type', 'text/plain')
                self.end_headers()
                self.wfile.write(res)
                self.wfile.write("\n")
            else:
                self.log_error ("No validation result to '%s' (responding 403)" % (self.path))
                self.send_response(403, 'Forbidden')
                self.end_headers()
        else:
            self.log_error ("Bad URL '%s' - I'm serving '%s' (responding 403)" % (self.path, args.serve_url))
            self.send_response(403, 'Forbidden')
            self.end_headers()

    def log_error(self, fmt, *fmt_args):
        """ Log to syslog. """
        msg = self.my_address_string() + " - - " + fmt % fmt_args
        my_log_message(args, syslog.LOG_ERR, msg)

    def log_message(self, fmt, *fmt_args):
        """ Log to syslog. """
        msg = self.my_address_string() + " - - " + fmt % fmt_args
        my_log_message(args, syslog.LOG_INFO, msg)

    def my_address_string(self):
        """ For logging client host without resolving. """
        return self.client_address[0]

class YHSM_VALServer(BaseHTTPServer.HTTPServer):
    """
    Wrapper class to properly initialize address_family for IPv6 addresses.
    """
    def __init__(self, server_address, req_handler):
        if ":" in server_address[0]:
            self.address_family = socket.AF_INET6
        BaseHTTPServer.HTTPServer.__init__(self, server_address, req_handler)


def validate_yubikey_otp_short(self, params):
    """
    Validate YubiKey OTP using YubiHSM internal database.
    """
    from_key = params["otp"][0]
    if not re.match(ykotp_valid_input, from_key):
        self.log_error("IN: %s, Invalid OTP" % (from_key))
        return "ERR Invalid OTP"
    try:
        res = pyhsm.yubikey.validate_otp(hsm, from_key)
        return "OK counter=%04x low=%04x high=%02x use=%02x" % \
            (res.use_ctr, res.ts_low, res.ts_high, res.session_ctr)
    except pyhsm.exception.YHSM_CommandFailed, e:
        return "ERR %s" % (pyhsm.defines.status2str(e.status))

def validate_yubikey_otp(self, params):
    """
    Validate YubiKey OTP using YubiHSM internal database.
    """
    vres = {}
    from_key = params["otp"][0]
    if not re.match(ykotp_valid_input, from_key):
        self.log_error("IN: %s, Invalid OTP" % (from_key))
        vres["status"] = "BAD_OTP"
    else:
        vres["otp"] = from_key
    if not "nonce" in params:
        self.log_error("IN: %s, no nonce" % (from_key))
        vres["status"] = "MISSING_PARAMETER"
    else:
        nonce = params["nonce"][0]
        if len(nonce) < 16 or len(nonce) > 40:
            self.log_error("IN: %s, bad nonce : %s" % (from_key, nonce))
            vres["status"] = "MISSING_PARAMETER"
        else:
            vres["nonce"] = nonce
    if "sl" in params and not (params["sl"] == "100" or params["sl"] == "secure"):
        self.log_error("IN: %s, sync level unsupported" % (from_key))
        vres["status"] = "BACKEND_ERROR"
    sig, client_key, = check_signature(params)
    if sig != True:
        self.log_error("IN: %s, signature validation error" % (from_key))
        if client_key == None:
            vres["status"] = "NO_SUCH_CLIENT"
            # To be compatible with YK-VAL, we will sign this response using a null key
            client_key = chr(0)
        else:
            vres["status"] = "BAD_SIGNATURE"
    if "status" not in vres:
        try:
            res = pyhsm.yubikey.validate_otp(hsm, from_key)
            vres.update({"status": "OK",
                         "sessioncounter": str(res.use_ctr), # known confusion
                         "sessionuse": str(res.session_ctr), # known confusion
                         "timestamp": str((res.ts_high << 16 | res.ts_low) / 8)
                         }
                        )
            if "sl" in params:
                vres["sl"] = "100"
                if "timestamp" in params:
                    vres["t"] = time.strftime("%FT%TZ0000", time.gmtime())
        except pyhsm.exception.YHSM_CommandFailed, e:
            if e.status == pyhsm.defines.YSM_ID_NOT_FOUND:
                vres["status"] = "BAD_OTP"
            elif e.status == pyhsm.defines.YSM_OTP_REPLAY:
                vres["status"] = "REPLAYED_OTP"
            elif e.status == pyhsm.defines.YSM_OTP_INVALID:
                vres["status"] = "BAD_OTP"
            else:
                vres["status"] = "BACKEND_ERROR"
            self.log_error("IN: %s, validation result %s (replying %s)" % (from_key, pyhsm.defines.status2str(e.status), vres["status"]))

    return make_otp_response(vres, client_key)

def make_otp_response(vres, client_key):
    """
    Create validation response (signed, if a client key is supplied).
    """
    if client_key is not None:
        sig = make_signature(vres, client_key)
        vres['h'] = sig
    # produce "key=value" pairs from vres
    pairs = [x + "=" + ''.join(vres[x]) for x in sorted(vres.keys())]
    return '\n'.join(pairs)

def check_signature(params):
    """
    Verify the signature of the parameters in an OTP v2.0 verify request.

    Returns ValResultBool, Key
    """
    if 'id' in params:
        try:
            id_int = int(params['id'][0])
        except:
            my_log_message(args, syslog.LOG_INFO, "Non-numerical client id (%s) in request." % (params['id'][0]))
            return False, None
        key = client_ids.get(id_int)
        if key:
            if 'h' in params:
                sig = params['h'][0]
                good_sig = make_signature(params, key)
                if sig == good_sig:
                    #my_log_message(args, syslog.LOG_DEBUG, "Good signature (client id '%i')" % id_int)
                    return True, key
                else:
                    my_log_message(args, syslog.LOG_INFO, "Bad signature from client id '%i' (%s, expected %s)." \
                                       % (id_int, sig, good_sig))
            else:
                my_log_message(args, syslog.LOG_INFO, "Client id (%i) but no HMAC in request." % (id_int))
                return False, key
        else:
            my_log_message(args, syslog.LOG_INFO, "Unknown client id '%i'" % (id_int))
            return False, None
    return True, None

def make_signature(params, hmac_key):
    """
    Calculate a HMAC-SHA-1 (using hmac_key) of all the params except "h=".

    Returns base64 encoded signature as string.
    """
    # produce a list of "key=value" for all entries in params except `h'
    pairs = [x + "=" + ''.join(params[x]) for x in sorted(params.keys()) if x != "h"]
    sha = hmac.new(hmac_key, '&'.join(pairs), hashlib.sha1)
    return base64.b64encode(sha.digest())

def validate_oath_hotp(self, params):
    """
    Validate OATH-HOTP code using YubiHSM HMAC-SHA1 hashing with token keys
    secured in AEAD's that we have stored in an SQLite3 database.
    """
    from_key = params["hotp"][0]
    if not re.match(hotp_valid_input, from_key):
        self.log_error("IN: %s, Invalid OATH-HOTP OTP" % (params))
        return "ERR Invalid OATH-HOTP OTP"
    uid, otp, = get_oath_hotp_bits(params)
    if not uid or not otp:
        self.log_error("IN: %s, could not get UID/OTP ('%s'/'%s')" % (params, uid, otp))
        return "ERR Invalid OATH-HOTP input"
    if args.debug:
        print "OATH-HOTP uid %s, OTP %s" % (uid, otp)

    # Fetch counter value for `uid' from database
    try:
        db = ValOathDb(args.db_file)
        entry = db.get(uid)
    except Exception, e:
        self.log_error("IN: %s, database error : '%s'" % (params, e))
        return "ERR Internal error"

    # Check for correct OATH-HOTP OTP
    nonce = entry.data["nonce"].decode('hex')
    aead = entry.data["aead"].decode('hex')
    new_counter = pyhsm.oath_hotp.search_for_oath_code(hsm, entry.data["key_handle"], nonce, aead, \
                                                           entry.data["oath_c"], otp, args.look_ahead)
    if args.debug:
        print "OATH-HOTP %i..%i -> new C == %s" \
            % (entry.data["oath_c"], entry.data["oath_c"] + args.look_ahead, new_counter)
    if type(new_counter) != int:
        # XXX increase 'throttling parameter' to make brute forcing harder/impossible
        return "ERR Could not validate OATH-HOTP OTP"
    try:
        # Must successfully store new_counter before we return OK
        if db.update_oath_hotp_c(entry, new_counter):
            return "OK counter=%04x" % (new_counter)
        else:
            return "ERR replayed OATH-HOTP"
    except Exception, e:
        self.log_error("IN: %s, database error updating counter : %s" % (params, e))
        return "ERR Internal error"

def validate_oath_totp(self, params):
    """
    Validate OATH-TOTP code using YubiHSM HMAC-SHA1 hashing with token keys
    secured in AEAD's that we have stored in an SQLite3 database.
    """
    from_key = params["totp"][0]
    if not re.match(totp_valid_input, from_key):
        self.log_error("IN: %s, Invalid OATH-TOTP OTP" % (params))
        return "ERR Invalid OATH-TOTP OTP"
    uid, otp, = get_oath_totp_bits(params)
    if not uid or not otp:
        self.log_error("IN: %s, could not get UID/OTP ('%s'/'%s')" % (params, uid, otp))
        return "ERR Invalid OATH-TOTP input"
    if args.debug:
        print "OATH-TOTP uid %s, OTP %s" % (uid, otp)

    # Fetch counter value for `uid' from database
    try:
        db = ValOathDb(args.db_file)
        entry = db.get(uid)
    except Exception, e:
        self.log_error("IN: %s, database error : '%s'" % (params, e))
        return "ERR Internal error"

    # Check for correct OATH-TOTP OTP
    nonce = entry.data["nonce"].decode('hex')
    aead = entry.data["aead"].decode('hex')
    new_timecounter = pyhsm.oath_totp.search_for_oath_code(
        hsm, entry.data["key_handle"], nonce, aead, otp, args.interval, args.tolerance)

    if args.debug:
        print "OATH-TOTP counter: %i, interval: %i -> new timecounter == %s" \
             % (entry.data["oath_c"], args.interval, new_timecounter)
    if type(new_timecounter) != int:
        return "ERR Could not validate OATH-TOTP OTP"
    try:
        # Must successfully store new_timecounter before we return OK
        # Can use existing hotp function since it would be identical
        if db.update_oath_hotp_c(entry, new_timecounter):
            return "OK timecounter=%04x" % (new_timecounter)
        else:
            return "ERR replayed OATH-TOTP"
    except Exception, e:
        self.log_error("IN: %s, database error updating counter : %s" % (params, e))
        return "ERR Internal error"

def validate_pwhash(_self, params):
    """
    Validate password hash using YubiHSM.
    """
    pwhash, nonce, aead, key_handle = get_pwhash_bits(params)
    d_aead = aead.decode('hex')
    plaintext_len = len(d_aead) - pyhsm.defines.YSM_AEAD_MAC_SIZE
    pw = pwhash.ljust(plaintext_len, chr(0x0))
    if hsm.validate_aead(nonce.decode('hex'), key_handle, d_aead, pw):
        return "OK pwhash validated"
    return "ERR Could not validate pwhash"

def get_pwhash_bits(params):
    """ Extract bits for password hash validation from params. """
    if not "pwhash" in params or \
            not "nonce" in params or \
            not "aead" in params or \
            not "kh" in params:
        raise Exception("Missing required parameter in request (pwhash, nonce, aead or kh)")
    pwhash = params["pwhash"][0]
    nonce = params["nonce"][0]
    aead = params["aead"][0]
    key_handle = pyhsm.util.key_handle_to_int(params["kh"][0])
    return pwhash, nonce, aead, key_handle

def get_oath_hotp_bits(params):
    """ Extract the OATH-HOTP uid and OTP from params. """
    if "uid" in params:
        return params["uid"][0], int(params["hotp"][0])
    m = re.match("^([cbdefghijklnrtuv]*)([0-9]{6,8})", params["hotp"][0])
    uid, otp, = m.groups()
    return uid, int(otp),

def get_oath_totp_bits(params):
    """ Extract the OATH-TOTP uid and OTP from params. """
    if "uid" in params:
        return params["uid"][0], int(params["totp"][0])
    m = re.match("^([cbdefghijklnrtuv]*)([0-9]{6,8})", params["totp"][0])
    uid, otp, = m.groups()
    return uid, int(otp),

class ValOathDb():
    """ Provides access to database with AEAD's and other information for OATH tokens. """

    def __init__(self, filename):
        self.filename = filename
        self.conn = sqlite3.connect(self.filename)
        self.conn.row_factory = sqlite3.Row

    def get(self, key):
        """ Fetch entry from database. """
        c = self.conn.cursor()
        for row in c.execute("SELECT key, nonce, key_handle, aead, oath_C, oath_T FROM oath WHERE key = ?", (key,)):
            return ValOathEntry(row)
        raise Exception("OATH token for '%s' not found in database (%s)" % (key, self.filename))

    def update_oath_hotp_c(self, entry, new_c):
        """
        Update the OATH-HOTP counter value for `entry' in the database.

        Use SQL statement to ensure we only ever increase the counter.
        """
        key = entry.data["key"]
        c = self.conn.cursor()
        c.execute("UPDATE oath SET oath_c = ? WHERE key = ? AND ? > oath_c",
                  (new_c, key, new_c,))
        self.conn.commit()
        return c.rowcount == 1

class ValOathEntry():
    """ Class to hold a row of ValOathDb. """
    def __init__(self, row):
        if row:
            self.data = row

def parse_args():
    """
    Parse the command line arguments
    """
    parser = argparse.ArgumentParser(description = "Validate secrets using YubiHSM",
                                     add_help=True,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     )
    parser.add_argument('-D', '--device',
                        dest='device',
                        default=default_device,
                        required=False,
                        help='YubiHSM device',
                        )
    parser.add_argument('-U', '--serve-url',
                        dest='serve_url',
                        default=default_serve_url,
                        required=False,
                        help='Base URL for validation web service',
                        )
    parser.add_argument('-v', '--verbose',
                        dest='verbose',
                        action='store_true', default=False,
                        help='Enable verbose operation',
                        )
    parser.add_argument('--debug',
                        dest='debug',
                        action='store_true', default=False,
                        help='Enable debug operation',
                        )
    parser.add_argument('--port',
                        dest='listen_port',
                        type=int, default=8003,
                        required=False,
                        help='Port to listen on',
                        metavar='PORT',
                        )
    parser.add_argument('--addr',
                        dest='listen_addr',
                        default="127.0.0.1",
                        required=False,
                        help='Address to bind to',
                        metavar='ADDR',
                        )
    parser.add_argument('--hmac-kh',
                        dest='hmac_kh',
                        required=False, default=0,
                        help='Key handle to use for creating HMAC-SHA1 hashes',
                        metavar='KEY_HANDLE',
                        )
    parser.add_argument('--short-otp',
                        dest='mode_short_otp',
                        action='store_true', default=False,
                        help='Enable YubiKey OTP validation (KSM style response)',
                        )
    parser.add_argument('--otp',
                        dest='mode_otp',
                        action='store_true', default=False,
                        help='Enable YubiKey OTP validation 2.0',
                        )
    parser.add_argument('--hotp',
                        dest='mode_hotp',
                        action='store_true', default=False,
                        help='Enable OATH-HOTP validation',
                        )
    parser.add_argument('--totp',
                        dest='mode_totp',
                        action='store_true', default=False,
                        help='Enable OATH-TOTP validation',
                        )
    parser.add_argument('--pwhash',
                        dest='mode_pwhash',
                        action='store_true', default=False,
                        help='Enable password hash validation',
                        )
    parser.add_argument('--db-file',
                        dest='db_file',
                        default=default_db_file,
                        required=False,
                        help='DB file for storing AEAD\'s etc. for --pwhash and --hotp',
                        metavar='FILENAME',
                        )
    # XXX bad interaction with argparse.ArgumentDefaultsHelpFormatter here - we don't want to
    # use default=default_clients_file since we need to know if this option was specified explicitly
    # or not.
    parser.add_argument('--clients-file',
                        dest='clients_file',
                        default=None,
                        required=False,
                        help='File with OTP validation clients shared secrets. for --otp. Default : %s' % (default_clients_file),
                        metavar='FILENAME',
                        )
    parser.add_argument('--hotp-window',
                        dest='look_ahead',
                        type=int, required=False,
                        default = default_hotp_window,
                        help='Number of OATH-HOTP codes to search',
                        metavar='NUM',
                        )
    parser.add_argument('--totp-interval',
                        dest='interval',
                        type=int, required=False,
                        default = default_totp_interval,
                        help='Timeframe in seconds for a valid OATH-TOTP code',
                        metavar='NUM',
                        )
    parser.add_argument('--totp-tolerance',
                        dest='tolerance',
                        type=int, required=False,
                        default = default_totp_tolerance,
                        help='Tolerance in time-steps for a valid OATH-TOTP code',
                        metavar='NUM',
                        )
    parser.add_argument('--pid-file',
                        dest='pid_file',
                        default=default_pid_file,
                        required=False,
                        help='PID file',
                        metavar='FILENAME',
                        )

    return parser.parse_args()

def args_fixup():
    """
    Various cleanups/initializations based on result of parse_args().
    """
    global saved_key_handle
    saved_key_handle = args.hmac_kh

    args.key_handle = pyhsm.util.key_handle_to_int(args.hmac_kh)

    if not (args.mode_otp or args.mode_short_otp or args.mode_totp or args.mode_hotp or args.mode_pwhash):
        my_log_message(args, syslog.LOG_ERR, 'No validation mode enabled')
        sys.exit(1)

    global client_ids
    if args.clients_file != None:
        if not args.mode_otp:
            my_log_message(args, syslog.LOG_ERR, 'Clients file should only be used with --otp.')
            sys.exit(1)
        client_ids = load_clients_file(args.clients_file)
        if not client_ids:
            my_log_message(args, syslog.LOG_ERR, 'Failed loading clients file "%s"' % (args.clients_file))
            sys.exit(1)
    else:
        # we accept failure to load this file when the default is used
        loaded_client_ids = load_clients_file(default_clients_file)
        if loaded_client_ids:
            args.clients_file = default_clients_file
            client_ids = loaded_client_ids

def load_clients_file(filename):
    """
    Load a list of base64 encoded shared secrets for numerical client ids.

    Returns a dict.

    Format of file is expected to be

        # This is a comment. Blank lines are OK.

        123,c2hhcmVkIHNlY3JldA==
        456,MTIzNDU2Nzg5MDEyMw==
    """
    res = {}
    content = []
    try:
        fhandle = file(filename)
        content = fhandle.readlines()
        fhandle.close()
    except IOError:
        return None
    linenum = 0
    for line in content:
        linenum += 1
        while line.endswith("\r") or line.endswith("\n"):
            line = line[:-1]
        if re.match("(^\s*#|^\s*$)", line):
            # skip comments and empty lines
            continue
        parts = [x.strip() for x in line.split(',')]
        try:
            if len(parts) != 2:
                raise Exception()
            id_num = int(parts[0])
            key = base64.b64decode(parts[1])
            res[id_num] = key
        except:
            my_log_message(args, syslog.LOG_ERR, 'Bad data on line %i of clients file "%s" : "%s"' % (linenum, filename, line))
            return None
    return res

def write_pid_file(fn):
    """ Create a file with our PID. """
    if not fn:
        return None
    if fn == '' or fn == "''":
        # work around argument passings in init-scripts
        return None
    f = open(fn, "w")
    f.write("%s\n" % (os.getpid()))
    f.close()

def run():
    """
    Start the BaseHTTPServer and serve requests forever.
    """
    server_address = (args.listen_addr, args.listen_port)
    httpd = YHSM_VALServer(server_address, YHSM_VALRequestHandler)
    my_log_message(args, syslog.LOG_INFO, "Serving requests to 'http://%s:%s%s' (YubiHSM: '%s')" \
                       % (args.listen_addr, args.listen_port, args.serve_url, args.device))
    httpd.serve_forever()

def my_log_message(my_args, prio, msg):
    """
    Log msg to syslog, and possibly also output to stderr.
    """
    syslog.syslog(prio, msg)
    if my_args.debug or my_args.verbose or prio == syslog.LOG_ERR:
        sys.stderr.write("%s\n" % (msg))

def main():
    """
    The main function that will be executed when running this as a stand alone script.
    """
    my_name = os.path.basename(sys.argv[0])
    if not my_name:
        my_name = "yhsm-validation-server"
    syslog.openlog(my_name, syslog.LOG_PID, syslog.LOG_LOCAL0)

    global args
    args = parse_args()
    args_fixup()

    global hsm
    try:
        hsm = pyhsm.YHSM(device = args.device, debug = args.debug)
    except serial.SerialException, e:
        my_log_message(args, syslog.LOG_ERR, 'Failed opening YubiHSM device "%s" : %s' %(args.device, e))
        return 1

    write_pid_file(args.pid_file)

    try:
        run()
    except KeyboardInterrupt:
        print ""
        print "Shutting down"
        print ""


if __name__ == '__main__':
    sys.exit(main())



#
# Copyright (c) 2013-2014 Yubico AB
# See the file COPYING for licence statement.
#
"""
Export AEAD from database.
"""

import os
import sys
import errno
import argparse
import sqlalchemy


import pyhsm.aead_cmd


def insert_slash(string, every=2):
    """insert_slash insert / every 2 char"""
    return os.path.join(string[i:i+every] for i in xrange(0, len(string), every))


def mkdir_p(path):
    """mkdir -p: creates path like mkdir -p"""
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else: raise


def main():
    parser = argparse.ArgumentParser(description='Import AEADs into the database')

    parser.add_argument('path', help='filesystem path of where to put AEADs')
    parser.add_argument('dburl', help='connection URL for the database')
    args = parser.parse_args()


    #set the path
    path = args.path
    if not os.path.isdir(path):
        print("\nInvalid path, make sure it exists.\n")
        return 2

    #mysql url
    databaseUrl = args.dburl

    try:
        #check database connection
        engine = sqlalchemy.create_engine(databaseUrl)

        #SQLAlchemy voodoo
        metadata = sqlalchemy.MetaData()
        aeadobj = sqlalchemy.Table('aead_table', metadata, autoload=True, autoload_with=engine)
        connection = engine.connect()

    except:
        print("FATAL: Database connect failure")
        return 1

    aead = None
    nonce = None
    key_handle = None

    aead = pyhsm.aead_cmd.YHSM_GeneratedAEAD(nonce, key_handle, aead)

    #get data from the database
    result = connection.execute("SELECT * from aead_table")

    #cycle through resutls
    for row in result:
        #read values row by row
        aead.data = row['aead']
        publicId = row['public_id']
        aead.key_handle = row['keyhandle']
        aead.nonce = row['nonce']

        aead_dir = os.path.join(path, str(hex(aead.key_handle)).rstrip('L'), insert_slash(publicId))
        #sanitize path
        aead_dir = os.path.normpath(aead_dir)
        #create path
        mkdir_p(aead_dir)

        #write the file in the path
        pyhsm.aead_cmd.YHSM_GeneratedAEAD.save(aead, os.path.join(aead_dir, publicId))

    #close connection
    connection.close()


if __name__ == '__main__':
    sys.exit(main())



#
# Copyright (C) 2013 Yubico AB. All rights reserved.
#
"""
This is a daemon to allow multiple users of a YubiHSM without requiring
permission to use the device. The daemon listens on a TCP port on localhost
and allows multiple connections to share a YubiHSM. Access the YubiHSM via
the daemon by specifying a device string following the yhsm://<host>:<port>
syntax:

hsm = YHSM('yhsm://localhost:5348')

Note that the daemon and clients need to share the same version number to be
compatible.
"""

import sys
import socket
import json
import threading
import argparse
import pyhsm.stick
import daemon
import os


CMD_WRITE = 0
CMD_READ = 1
CMD_FLUSH = 2
CMD_DRAIN = 3
CMD_LOCK = 4
CMD_UNLOCK = 5

COMMANDS = {
    CMD_WRITE: 'write',
    CMD_READ: 'read',
    CMD_FLUSH: 'flush',
    CMD_DRAIN: 'drain'
}

context = daemon.DaemonContext()

def pack_data(data):
    if isinstance(data, basestring):
        return data.encode('base64')
    return data


def unpack_data(data):
    if isinstance(data, basestring):
        return data.decode('base64')
    return data

def write_pid_file(fn):
    """ Create a file with our PID. """
    if not fn:
        return None
    if fn == '' or fn == "''":
        # work around argument passings in init-scripts
        return None
    f = open(fn, "w")
    f.write("%s\n" % (os.getpid()))
    f.close()

class YHSM_Stick_Server():
    def __init__(self, device, addr):
        self.device = device
        self._stick = None
        self.pidfile = None

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self.socket.bind(addr)
        self.lock = threading.Lock()

    def serve(self):
        write_pid_file(self.pidfile)

        self.socket.listen(20)

        try:
            while True:
                cs, address = self.socket.accept()
                thread = threading.Thread(target=self.client_handler,
                                          args=(cs,))
                thread.start()
        except Exception as e:
            print e
            sys.exit(1)

    def invoke(self, cmd, *args):
        try:
            if not self._stick:
                self._stick = pyhsm.stick.YHSM_Stick(self.device)
            res = getattr(self._stick, COMMANDS[cmd])(*args)
        except Exception as e:
            res = e
            print e
            self._stick = None
        return res

    def client_handler(self, socket):
        socket_file = socket.makefile('wb')
        has_lock = False

        try:
            while True:
                data = json.loads(socket_file.readline())
                cmd = data[0]
                args = map(unpack_data, data[1:])
                if cmd == CMD_LOCK:
                    if not has_lock:
                        self.lock.acquire()
                        has_lock = True
                elif has_lock:
                    if cmd == CMD_UNLOCK:
                        self.lock.release()
                        has_lock = False
                    else:
                        res = self.invoke(cmd, *args)
                        json.dump(pack_data(res), socket_file)
                        socket_file.write("\n")
                        socket_file.flush()
                else:
                    err = 'Command run without holding lock!'
                    print err
                    json.dump({'error': err}, socket_file)
                    socket_file.write("\n")
                    socket_file.flush()
                    break
        except Exception:
            # Client disconnected, ignore.
            pass
        finally:
            if has_lock:
                self.lock.release()
            socket_file.close()
            socket.close()


def main():
    parser = argparse.ArgumentParser(
        description='YubiHSM server daemon',
        add_help=True,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('-D', '--device', nargs='?', default='/dev/ttyACM0',
                        help='device name')
    parser.add_argument('-d', '--daemon', default=False,
                        action='store_true', help='run as daemon')
    parser.add_argument('-I', '--interface', nargs='?', default='localhost',
                        help='network interface to bind to')
    parser.add_argument('-P', '--port', nargs='?', type=int, default=5348,
                        help='TCP port to bind to')
    parser.add_argument('--pid-file',
                        dest='pid_file',
                        default=None,
                        required=False,
                        help='PID file',
                        metavar='FILENAME')

    args = parser.parse_args()

    print 'Starting YubiHSM daemon for device: %s, listening on: %s:%d' % \
        (args.device, args.interface, args.port)

    server = YHSM_Stick_Server(args.device, (args.interface, args.port))
    print 'You can connect to the server using the following device string:'
    print 'yhsm://127.0.0.1:%d' % args.port

    server.pidfile = args.pid_file
    context.files_preserve = [server.socket]
    if args.daemon:
        with context:
            server.serve()
    else:
        server.serve()



"""
helper functions to work with OATH TOTP (RFC6238) OTP's and YubiHSM
"""

# Copyright (c) 2016 Storedsafe AB
# Copyright (c) 2011 Yubico AB
# See the file COPYING for licence statement.

import datetime
import time
from pyhsm.oath_hotp import search_for_oath_code as search_hotp

__all__ = [
    # constants
    # functions
    'search_for_oath_code',
    # classes
]


def search_for_oath_code(hsm, key_handle, nonce, aead, user_code, interval=30,
                         tolerance=0):
    """
    Try to validate an OATH TOTP OTP generated by a token whose secret key is
    available to the YubiHSM through the AEAD.

    The parameter `aead' is either a string, or an instance of YHSM_GeneratedAEAD.

    Returns timecounter value on successful auth, and None otherwise.
    """

    # timecounter is the lowest acceptable value based on tolerance
    timecounter = timecode(datetime.datetime.now(), interval) - tolerance

    return search_hotp(
        hsm, key_handle, nonce, aead, timecounter, user_code, 1 + 2*tolerance)


def timecode(time_now, interval):
    """ make integer and divide by time interval of valid OTP """
    i = time.mktime(time_now.timetuple())
    return int(i / interval)



#
# Tool to validate a YubiKey OTP using the YubiHSM internal database.
#
# This requires that you have imported the secret AES key of the YubiKey
# into the database with `../yubikey-ksm/yhsm-import-keys --internal-db'
# or otherwise.
#
# Copyright (c) 2011 Yubico AB
# See the file COPYING for licence statement.
#

import os
import re
import sys
import struct
import argparse
import pyhsm
import pyhsm.yubikey

default_device = "/dev/ttyACM0"

def parse_args():
    """
    Parse the command line arguments
    """
    global default_device

    parser = argparse.ArgumentParser(description = "Validate YubiKey OTP's using YubiHSM",
                                     add_help=True
                                     )
    parser.add_argument('-D', '--device',
                        dest='device',
                        default=default_device,
                        required=False,
                        help='YubiHSM device (default : %s).' % default_device
                        )
    parser.add_argument('-v', '--verbose',
                        dest='verbose',
                        action='store_true', default=False,
                        help='Enable verbose operation.'
                        )
    parser.add_argument('--debug',
                        dest='debug',
                        action='store_true', default=False,
                        help='Enable debug operation.'
                        )
    group = parser.add_argument_group('Modes', 'What you want to validate')
    mode_group = group.add_mutually_exclusive_group(required = True)
    mode_group.add_argument('--otp',
                            dest='otp',
                            help='The output from your YubiKey.'
                            )
    mode_group.add_argument('--oath',
                            dest='oath',
                            help='The output from your OATH-HOTP token.'
                            )

    args = parser.parse_args()
    return args

def validate_otp(hsm, args):
    """
    Validate an OTP.
    """
    try:
        res = pyhsm.yubikey.validate_otp(hsm, args.otp)
        if args.verbose:
            print "OK counter=%04x low=%04x high=%02x use=%02x" % \
                (res.use_ctr, res.ts_low, res.ts_high, res.session_ctr)
        return 0
    except pyhsm.exception.YHSM_CommandFailed, e:
        if args.verbose:
            print "%s" % (pyhsm.defines.status2str(e.status))
        # figure out numerical response code
        for r in [pyhsm.defines.YSM_OTP_INVALID, \
                      pyhsm.defines.YSM_OTP_REPLAY, \
                      pyhsm.defines.YSM_ID_NOT_FOUND]:
            if e.status == r:
                return r - pyhsm.defines.YSM_RESPONSE
        # not found
        return 0xff

def validate_oath(hsm, args):
    """
    Validate an OATH OTP.
    """
    print "ERROR: Not implemented, try 'yhsm-validation-server'."
    return 0


def main():
    args = parse_args()

    if args.debug:
        print "YHSM device		: %s" % (args.device)
        print ""

    hsm = pyhsm.YHSM(device = args.device, debug=args.debug)

    status = 1
    if args.otp:
        status = validate_otp(hsm, args)
    elif args.oath:
        status = validate_oath(hsm, args)

    return status


if __name__ == '__main__':
    sys.exit(main())



#
# Copyright (c) 2013-2014 Yubico AB
# See the file COPYING for licence statement.
#
"""
Import AEADs to database.
"""

import os
import re
import sys
import argparse
import sqlalchemy

from pyhsm.util import key_handle_to_int
import pyhsm.aead_cmd


def extract_keyhandle(path, filepath):
    """extract keyhandle value from the path"""

    keyhandle = filepath.lstrip(path)
    keyhandle = keyhandle.split("/")
    return keyhandle[0]


def insert_query(connection, publicId, aead, keyhandle, aeadobj):
    """this functions read the response fields and creates sql query. then
    inserts everything inside the database"""

    # turn the keyhandle into an integer
    keyhandle = key_handle_to_int(keyhandle)
    if not keyhandle == aead.key_handle:
        print("WARNING: keyhandle does not match aead.key_handle")
        return None

    # creates the query object
    try:
        sql = aeadobj.insert().values(public_id=publicId, keyhandle=aead.key_handle, nonce=aead.nonce, aead=aead.data)
        # insert the query
        result = connection.execute(sql)
        return result
    except sqlalchemy.exc.IntegrityError:
        pass
    return None


def main():
    parser = argparse.ArgumentParser(description='Import AEADs into the database')

    parser.add_argument('path', help='filesystem path of where to find AEADs')
    parser.add_argument('dburl', help='connection URL for the database')

    args = parser.parse_args()

    path = args.path
    databaseUrl = args.dburl

    if not os.path.isdir(path):
        print("\nInvalid path, check your spelling.\n")
        return 2

    try:
        engine = sqlalchemy.create_engine(databaseUrl)

        #SQLAlchemy voodoo
        metadata = sqlalchemy.MetaData()
        aeadobj = sqlalchemy.Table('aead_table', metadata, autoload=True, autoload_with=engine)
        connection = engine.connect()
    except:
        print("FATAL: Database connect failure")
        return 1

    for root, subFolders, files in os.walk(path):
        if files:
            if not re.match(r'^[cbdefghijklnrtuv]+$', files[0]):
                continue

            #build file path
            filepath = os.path.join(root, files[0])

            #extract the key handle from the path
            keyhandle = extract_keyhandle(path, filepath)
            kh_int = pyhsm.util.key_handle_to_int(keyhandle)

            #instantiate a new aead object
            aead = pyhsm.aead_cmd.YHSM_GeneratedAEAD(None, kh_int, '')
            aead.load(filepath)

            #set the public_id
            public_id = str(files[0])

            #check it is old format aead
            if not aead.nonce:
                #configure values for oldformat
                aead.nonce = pyhsm.yubikey.modhex_decode(public_id).decode('hex')
                aead.key_handle = key_handle_to_int(keyhandle)

            if not insert_query(connection, public_id, aead, keyhandle, aeadobj):
                print("WARNING: could not insert %s" % public_id)

    #close sqlalchemy
    connection.close()


if __name__ == '__main__':
    sys.exit(main())



#
# Tool to add an OATH token to the yhsm-validation-server database.
#
# Copyright (c) 2011 Yubico AB
# See the file COPYING for licence statement.
#

import sys
import struct
import sqlite3
import argparse
import pyhsm
import pyhsm.oath_hotp
from hashlib import sha1

default_device = "/dev/ttyACM0"
default_db_file = "/var/yubico/yhsm-validation-server.db"

def parse_args():
    """
    Parse the command line arguments
    """
    global default_device

    parser = argparse.ArgumentParser(description = 'Initialize OATH token for use with yhsm-validation-server',
                                     add_help=True,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     )
    parser.add_argument('-D', '--device',
                        dest='device',
                        default=default_device,
                        required=False,
                        help='YubiHSM device',
                        )
    parser.add_argument('-v', '--verbose',
                        dest='verbose',
                        action='store_true', default=False,
                        help='Enable verbose operation',
                        )
    parser.add_argument('--debug',
                        dest='debug',
                        action='store_true', default=False,
                        help='Enable debug operation',
                        )
    parser.add_argument('--force',
                        dest='force',
                        action='store_true', default=False,
                        help='Overwrite any present entry',
                        )
    parser.add_argument('--key-handle',
                        dest='key_handle',
                        required=True,
                        help='Key handle to create AEAD',
                        metavar='HANDLE',
                        )
    parser.add_argument('--uid',
                        dest='uid',
                        required=True,
                        help='User ID',
                        metavar='STR',
                        )
    parser.add_argument('--oath-c',
                        dest='oath_c',
                        required=False,
                        default = 0,
                        help='Initial OATH counter value',
                        metavar='INT',
                        )
    parser.add_argument('--test-oath-window',
                        dest='look_ahead',
                        required=False,
                        default = 10,
                        help='Number of codes to search with --test-code',
                        metavar='INT',
                        )
    parser.add_argument('--test-code',
                        dest='test_code',
                        type=int, required=False,
                        help='Optional OTP from token for verification',
                        metavar='INT',
                        )
    parser.add_argument('--oath-k',
                        dest='oath_k',
                        required=False,
                        help='The secret key of the token, hex encoded',
                        metavar='HEXSTR',
                        )
    parser.add_argument('--db-file',
                        dest='db_file',
                        default=default_db_file,
                        required=False,
                        help='DB file for storing AEAD\'s for --pwhash and --oath in the yhsm-validation-server',
                        metavar='FN',
                        )

    args = parser.parse_args()
    return args

def args_fixup(args):
    keyhandles_fixup(args)

def keyhandles_fixup(args):
    args.key_handle = pyhsm.util.key_handle_to_int(args.key_handle)

def generate_aead(hsm, args):
    """ Protect the oath-k in an AEAD. """
    key = get_oath_k(args)
    # Enabled flags 00010000 = YSM_HMAC_SHA1_GENERATE
    flags = struct.pack("< I", 0x10000)
    hsm.load_secret(key + flags)
    nonce = hsm.get_nonce().nonce
    aead = hsm.generate_aead(nonce, args.key_handle)
    if args.debug:
        print "AEAD: %s (%s)" % (aead.data.encode('hex'), aead)
    return nonce, aead

def validate_oath_c(hsm, args, nonce, aead):
    if args.test_code:
        if args.verbose:
            print "Trying to validate the OATH counter value in the range %i..%i." \
                % (args.oath_c, args.oath_c + args.look_ahead)
        counter = pyhsm.oath_hotp.search_for_oath_code(hsm, args.key_handle, nonce, aead, \
                                                           args.oath_c, args.test_code, args.look_ahead \
                                                           )
        if type(counter) != int:
            sys.stderr.write("Failed validating OTP %s (in range %i..%i) using supplied key.\n" \
                                 % (args.test_code, args.oath_c, args.oath_c + args.look_ahead))
            sys.exit(1)
        if args.verbose:
            print "OATH C==%i validated with code %s" % (counter - 1, args.test_code)
        return counter
    return args.oath_c

def get_oath_k(args):
    """ Get the OATH K value (secret key), either from args or by prompting. """
    if args.oath_k:
        decoded = args.oath_k.decode('hex')
    else:
        t = raw_input("Enter OATH key (hex encoded) : ")
        decoded = t.decode('hex')

    if len(decoded) > 20:
        decoded = sha1(decoded).digest()
    decoded = decoded.ljust(20, '\0')
    return decoded

class ValOathDb():
    """ Provides access to database with AEAD's and other information. """

    def __init__(self, filename):
        self.filename = filename
        self.conn = sqlite3.connect(self.filename)

        self.create_table()

    def create_table(self):
        c = self.conn.cursor()
        c.execute("CREATE TABLE IF NOT EXISTS oath " \
                      "(key TEXT PRIMARY KEY, nonce TEXT, key_handle INTEGER, aead TEXT, oath_C INTEGER, oath_T INTEGER)")

    def add(self, entry):
        """ Add entry to database. """
        c = self.conn.cursor()
        c.execute("INSERT INTO oath (key, aead, nonce, key_handle, oath_C, oath_T) VALUES (?, ?, ?, ?, ?, ?)",
                  (entry.data["key"], \
                       entry.data["aead"], \
                       entry.data["nonce"], \
                       entry.data["key_handle"], \
                       entry.data["oath_C"], \
                       entry.data["oath_T"],))
        self.conn.commit()
        return c.rowcount == 1

    def delete(self, entry):
        """ Delete entry from database. """
        c = self.conn.cursor()
        c.execute("DELETE FROM oath WHERE key = ?", (entry.data["key"],))

class ValOathEntry():
    """ Class to hold a row of ValOathDb. """
    def __init__(self, row):
        if row:
            self.data = row

def store_oath_entry(args, nonce, aead, oath_c):
    """ Store the AEAD in the database. """
    data = {"key": args.uid,
            "aead": aead.data.encode('hex'),
            "nonce": nonce.encode('hex'),
            "key_handle": args.key_handle,
            "oath_C": oath_c,
            "oath_T": None,
            }
    entry = ValOathEntry(data)
    db = ValOathDb(args.db_file)
    try:
        if args.force:
            db.delete(entry)
        db.add(entry)
    except sqlite3.IntegrityError, e:
        sys.stderr.write("ERROR: %s\n" % (e))
        return False
    return True

def main():
    args = parse_args()

    args_fixup(args)

    print "Key handle		: %s" % (args.key_handle)
    print "YHSM device		: %s" % (args.device)
    print ""

    hsm = pyhsm.YHSM(device = args.device, debug=args.debug)

    nonce, aead = generate_aead(hsm, args)
    oath_c = validate_oath_c(hsm, args, nonce, aead)
    if not store_oath_entry(args, nonce, aead, oath_c):
        return 1


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from setuptools import setup, find_packages
import re

VERSION_PATTERN = re.compile(r"(?m)^__version__\s*=\s*['\"](.+)['\"]$")


def get_version():
    """Return the current version as defined by yubico/yubico_version.py."""

    with open('pyhsm/__init__.py', 'r') as f:
        match = VERSION_PATTERN.search(f.read())
        return match.group(1)

setup(
    name='pyhsm',
    version=get_version(),
    description='Python code for talking to a YubiHSM',
    author='Dain Nilsson',
    author_email='dain@yubico.com',
    url='https://github.com/Yubico/python-pyhsm',
    license='BSD 2 clause',
    packages=find_packages(exclude=['test']),
    entry_points={
        'console_scripts': [
            # tools
            'yhsm-daemon = pyhsm.stick_daemon:main [daemon]',
            'yhsm-decrypt-aead = pyhsm.tools.decrypt_aead:main',
            'yhsm-generate-keys = pyhsm.tools.generate_keys:main',
            'yhsm-keystore-unlock = pyhsm.tools.keystore_unlock:main',
            'yhsm-linux-add-entropy = pyhsm.tools.linux_add_entropy:main',
            # ksm
            'yhsm-yubikey-ksm = pyhsm.ksm.yubikey_ksm:main [db,daemon]',
            'yhsm-import-keys = pyhsm.ksm.import_keys:main',
            'yhsm-db-export = pyhsm.ksm.db_export:main [db]',
            'yhsm-db-import = pyhsm.ksm.db_import:main [db]',
            # validation server
            'yhsm-validation-server = pyhsm.val.validation_server:main',
            'yhsm-validate-otp = pyhsm.val.validate_otp:main',
            'yhsm-init-oath-token = pyhsm.val.init_oath_token:main'
        ]
    },
    test_suite='test.test_init',
    tests_require=[],
    install_requires=[
        'pyserial >= 2.3',
        'pycrypto >= 2.1'
    ],
    extras_require={
        'db': ['sqlalchemy>=0.9.7'],
        'daemon': ['python-daemon']
    },
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
    ]
)



#
"""
Tool to generate YubiKey secret keys using YubiHSM.

After generation with this tool, you can (given that you know the AES
key for the key handle used in the HSM) generate a CSV file of the
unencrypted AEADs formatted for YubiKey personalization using the
YubiKey multi configuration utility (Windows) using the command
yhsm-decrypt-aead.

Example :

  1) Configure HSM with key handle 99 having key
      2000200020002000200020002000200020002000200020002000200020002000
  2) Generate 1000 AEADs for YubiKeys using something like this
     (XXXX can be a customer specific public_id prefix allocated by Yubico -
      0000-0009 (in modhex) are for tests)

      $ yhsm-generate-keys --key-handles 99 --start-public-id djXXXXcccccc \
            -O /var/cache/yubikey-ksm/aeads --count 1000
  3) Create CSV-file with
      $ yhsm-decrypt-aead --aes-key 2000...2000 --format yubikey-csv \
            /var/cache/yubikey-ksm/aeads
  4) Program YubiKeys using CSV file contents
  5) Start a KSM to decrypt OTPs from the YubiKeys
      $ yhsm-yubikey-ksm --key-handle 99
"""
#
# Copyright (c) 2011, 2012 Yubico AB
# See the file COPYING for licence statement.
#

import os
import sys
import argparse
import pyhsm
import pyhsm.yubikey

default_device = "/dev/ttyACM0"
default_dir = "/var/cache/yubikey-ksm/aeads"

def parse_args():
    """
    Parse the command line arguments
    """
    parser = argparse.ArgumentParser(description = "Generate secrets for YubiKeys using YubiHSM",
                                     add_help=True,
                                     formatter_class = argparse.ArgumentDefaultsHelpFormatter,
                                     )
    parser.add_argument('-D', '--device',
                        dest='device',
                        default=default_device,
                        required=False,
                        help='YubiHSM device',
                        )
    parser.add_argument('-O', '--output-dir', '--aead-dir',
                        dest='output_dir',
                        default=default_dir,
                        required=False,
                        help='Output directory (AEAD base dir)',
                        )
    parser.add_argument('-c', '--count',
                        dest='count',
                        type=int, default=1,
                        required=False,
                        help='Number of secrets to generate',
                        )
    parser.add_argument('-v', '--verbose',
                        dest='verbose',
                        action='store_true', default=False,
                        help='Enable verbose operation',
                        )
    parser.add_argument('--public-id-chars',
                        dest='public_id_chars',
                        type=int, default=12,
                        required=False,
                        help='Number of chars in generated public ids',
                        )
    parser.add_argument('--key-handles',
                        dest='key_handles',
                        nargs='+',
                        required=True,
                        help='Key handles to encrypt the generated secrets with',
                        )
    parser.add_argument('--start-public-id',
                        dest='start_id',
                        required=True,
                        help='The first public id to generate AEAD for',
                        )

    parser.add_argument('--random-nonce',
                        dest='random_nonce',
                        required=False,
                        action='store_true', default=False,
                        help='Let the HSM generate nonce',
                        )

    return parser.parse_args()

def args_fixup(args):
    if not os.path.isdir(args.output_dir):
        sys.stderr.write("Output directory '%s' does not exist.\n" % (args.output_dir))
        sys.exit(1)

    keyhandles_fixup(args)

    try:
        n = int(args.start_id)
    except ValueError:
        hexstr = pyhsm.yubikey.modhex_decode(args.start_id)
        n = int(hexstr, 16)

    if(n <= 0):
        sys.stderr.write("Start ID must be greater than 0, was %d\n" % (n))
        sys.exit(1)

    args.start_id = n


def keyhandles_fixup(args):
    """
    Walk through the supplied key handles and normalize them, while keeping
    the input format too (as value in a dictionary). The input format is
    used in AEAD filename paths.
    """
    new_handles = {}
    for val in args.key_handles:
        for this in val.split(','):
            n = pyhsm.util.key_handle_to_int(this)
            new_handles[n] = this

    args.key_handles = new_handles


def gen_keys(hsm, args):
    """
    The main key generating loop.
    """

    if args.verbose:
        print "Generating %i keys :\n" % (args.count)
    else:
        print "Generating %i keys" % (args.count)

    for int_id in range(args.start_id, args.start_id + args.count):
        public_id = ("%x" % int_id).rjust(args.public_id_chars, '0')
        padded_id = pyhsm.yubikey.modhex_encode(public_id)

        if args.verbose:
            print "  %s" % (padded_id)

        num_bytes = len(pyhsm.aead_cmd.YHSM_YubiKeySecret('a' * 16, 'b' * 6).pack())
        hsm.load_random(num_bytes)
        for kh in args.key_handles.keys():
            if args.random_nonce:
                nonce = ""
            else:
                nonce = public_id.decode('hex')
            aead = hsm.generate_aead(nonce, kh)

            filename = output_filename(args.output_dir, args.key_handles[kh], padded_id)

            if args.verbose:
                print "    %4s, %i bytes (%s) -> %s" % \
                    (args.key_handles[kh], len(aead.data), shorten_aead(aead), filename)
            aead.save(filename)

        if args.verbose:
            print ""

    print "\nDone\n"

def shorten_aead(aead):
    """ Produce pretty-printable version of long AEAD. """
    head = aead.data[:4].encode('hex')
    tail = aead.data[-4:].encode('hex')
    return "%s...%s" % (head, tail)

def output_filename(output_dir, key_handle, public_id):
    """
    Return an output filename for a generated AEAD. Creates a hashed directory structure
    using the last three bytes of the public id to get equal usage.
    """
    parts = [output_dir, key_handle] + pyhsm.util.group(public_id, 2)
    path = os.path.join(*parts)

    if not os.path.isdir(path):
        os.makedirs(path)

    return os.path.join(path, public_id)


def main():
    args = parse_args()

    args_fixup(args)

    print "output dir		: %s" % (args.output_dir)
    print "keys to generate	: %s" % (args.count)
    print "key handles		: %s" % (args.key_handles)
    print "start public_id		: %s (0x%x)" % (args.start_id, args.start_id)
    print "YHSM device		: %s" % (args.device)
    print ""

    if os.path.isfile(args.device):
        hsm = pyhsm.soft_hsm.SoftYHSM.from_file(args.device)
    else:
        hsm = pyhsm.YHSM(device=args.device)

    gen_keys(hsm, args)



#
"""
Tool to import YubiKey secrets to YubiHSM.

The default mode is to turn each YubiKey secret into an AEAD
(Authenticated Encryption with Associated Data) block that is stored
in a file on the host computer (one file per YubiKey). This enables
validation of virtually unlimited numbers of YubiKey's OTPs.

If --internal-db is used, the YubiKey secret will be stored inside
the YubiHSM, and complete validation (including counter management)
will be done inside the YubiHSM. The internal database is currently
limited to 1024 entries.

The input is supposed to be a comma-separated list of entries like this

  # ykksm 1
  123456,ftftftcccc,534543524554,fcacd309a20ce1809c2db257f0e8d6ea,000000000000,,,

  (seqno, public id, private uid, AES key, dunno,,,)

This is also the format of a database export from a traditional YK-KSM.
"""
#
# Copyright (c) 2011, 2012 Yubico AB
# See the file COPYING for licence statement.
#

import os
import sys
import argparse
import pyhsm
import pyhsm.yubikey
from pyhsm.soft_hsm import SoftYHSM

default_device = "/dev/ttyACM0"
default_dir = "/var/cache/yubikey-ksm/aeads"
default_public_id_chars = 12


def parse_args():
    """
    Parse the command line arguments
    """
    parser = argparse.ArgumentParser(
        description="Import existing secrets to YubiHSM eco system",
        add_help=True, formatter_class=argparse.ArgumentDefaultsHelpFormatter,)
    parser.add_argument('-D', '--device',
                        dest='device',
                        default=default_device,
                        required=False,
                        help='YubiHSM device',
                        )
    parser.add_argument('-O', '--output-dir', '--aead-dir',
                        dest='output_dir',
                        default=default_dir,
                        required=False,
                        help='Output directory (AEAD base dir)',
                        metavar='DIR',
                        )
    parser.add_argument('-v', '--verbose',
                        dest='verbose',
                        action='store_true', default=False,
                        help='Enable verbose operation',
                        )
    parser.add_argument('--debug',
                        dest='debug',
                        action='store_true', default=False,
                        help='Enable debug operation',
                        )
    parser.add_argument('--public-id-chars',
                        dest='public_id_chars',
                        type=int, default=default_public_id_chars,
                        required=False,
                        help='Number of chars in generated public ids',
                        metavar='NUM',
                        )
    parser.add_argument(
        '--key-handles',
        dest='key_handles',
        nargs='+',
        required=True,
        help='Key handle(s) to encrypt the imported secrets with',
        metavar='HANDLE',
     )
    parser.add_argument('--internal-db',
                        dest='internal_db',
                        action='store_true', default=False,
                        help='Store key in YubiHSM internal database',
                        )
    parser.add_argument(
        '--aes-key',
        dest='aes_key',
        required=False,
        default=None,
        help='AES key to use when generating AEADs (no YubiHSM)',
        metavar='HEXSTR',
     )
    parser.add_argument('--random-nonce',
                        dest='random_nonce',
                        action='store_true', default=False,
                        help='Let the HSM generate random nonce',
                        )

    args = parser.parse_args()
    if args.internal_db:
        if len(args.key_handles) != 1:
            sys.stderr.write("--internal-db requires exactly one key handle.\n")
            sys.exit(1)
        if args.aes_key:
            sys.stderr.write("--internal-db incompatible with --aes-key.\n")
            sys.exit(1)
    return args


def args_fixup(args):
    if not args.internal_db and not os.path.isdir(args.output_dir):
        sys.stderr.write(
            "Output directory '%s' does not exist.\n" %
            (args.output_dir))
        sys.exit(1)

    if args.aes_key:
        args.aes_key = args.aes_key.decode('hex')
    keyhandles_fixup(args)


def keyhandles_fixup(args):
    """
    Walk through the supplied key handles and normalize them, while keeping
    the input format too (as value in a dictionary). The input format is
    used in AEAD filename paths.
    """
    new_handles = {}
    for val in args.key_handles:
        n = pyhsm.util.key_handle_to_int(val)
        new_handles[n] = val

    args.key_handles = new_handles


def import_keys(hsm, args):
    """
    The main stdin iteration loop.
    """
    res = True

    # ykksm 1
    #123456,ftftftcccc,534543524554,fcacd309a20ce1809c2db257f0e8d6ea,000000000000,,,

    for line in sys.stdin:
        if line[0] == '#':
            continue

        l = line.split(',')
        modhex_id = l[1]
        uid = l[2].decode('hex')
        key = l[3].decode('hex')

        if modhex_id and uid and key:
            public_id = pyhsm.yubikey.modhex_decode(modhex_id)
            padded_id = modhex_id.rjust(args.public_id_chars, 'c')

        if int(public_id, 16) == 0:
            print "WARNING: Skipping import of key with public ID: %s" % (padded_id)
            print "This public ID is unsupported by the YubiHSM.\n"
            continue

        if args.verbose:
            print "  %s" % (padded_id)

        secret = pyhsm.aead_cmd.YHSM_YubiKeySecret(key, uid)
        hsm.load_secret(secret)

        for kh in args.key_handles.keys():
            if(args.random_nonce):
                nonce = ""
            else:
                nonce = public_id.decode('hex')
            aead = hsm.generate_aead(nonce, kh)

            if args.internal_db:
                if not store_in_internal_db(
                        args, hsm, modhex_id, public_id, kh, aead):
                    res = False
                continue

            filename = output_filename(
                args.output_dir, args.key_handles[kh], padded_id)

            if args.verbose:
                print "    %4s, %i bytes (%s) -> %s" % \
                    (args.key_handles[kh], len(aead.data), shorten_aead(aead), filename)

            aead.save(filename)

        if args.verbose:
            print ""

    if res:
        print "\nDone\n"
    else:
        print "\nDone (one or more entries rejected)"
    return res


def store_in_internal_db(args, hsm, modhex_id, public_id, kh, aead):
    """ Store record (AEAD) in YubiHSM internal DB """
    if args.verbose:
        print "    %i bytes (%s) -> internal db..." % \
            (len(aead.data), shorten_aead(aead)),
    try:
        hsm.db_store_yubikey(public_id.decode('hex'), kh, aead)
        if args.verbose:
            print "OK"
    except pyhsm.exception.YHSM_CommandFailed as e:
        if args.verbose:
            print "%s" % (pyhsm.defines.status2str(e.status))
        else:
            print "Storing ID %s FAILED: %s" % (modhex_id, pyhsm.defines.status2str(e.status))
        return False
    return True


def shorten_aead(aead):
    """ Produce pretty-printable version of long AEAD. """
    head = aead.data[:4].encode('hex')
    tail = aead.data[-4:].encode('hex')
    return "%s...%s" % (head, tail)


def output_filename(output_dir, key_handle, public_id):
    """
    Return an output filename for a generated AEAD. Creates a hashed directory structure
    using the last three bytes of the public id to get equal usage.
    """
    parts = [output_dir, key_handle] + pyhsm.util.group(public_id, 2)
    path = os.path.join(*parts)

    if not os.path.isdir(path):
        os.makedirs(path)

    return os.path.join(path, public_id)


def main():
    args = parse_args()

    args_fixup(args)

    if sys.stdin.readline() != "# ykksm 1\n":
        sys.stderr.write(
            "Did not get '# ykksm 1' header as first line of input.\n")
        sys.exit(1)

    print "output dir		: %s" % (args.output_dir)
    print "key handles		: %s" % (args.key_handles)
    print "YHSM device		: %s" % (args.device)
    print ""

    if args.aes_key:
        keys = {kh: args.aes_key for kh in args.key_handles}
        hsm = SoftYHSM(keys, args.debug)
    elif os.path.isfile(args.device):
        hsm = SoftYHSM.from_file(args.device, debug=args.debug)
    else:
        hsm = pyhsm.YHSM(device=args.device, debug=args.debug)

    return not import_keys(hsm, args)


if __name__ == '__main__':
    sys.exit(main())
